import rich_click as click
import os
from arkitekt_next.cli.constants import *
from getpass import getuser
from arkitekt_next.cli.types import Manifest, Requirement
from arkitekt_next.cli.vars import get_manifest, get_console
import semver
from arkitekt_next.cli.io import write_manifest
from rich.panel import Panel
from typing import Optional, List


def check_overwrite_app(ctx, param, value):
    """Callback to check and prompt for file overwrite."""

    app_file = ctx.params["entrypoint"] + ".py"
    if os.path.exists(app_file) and not value:
        should_overwrite = click.confirm(
            "Entrypoint File already exists. Do you want to overwrite?"
        )
        return should_overwrite

    return value


def check_overwrite(ctx, param, value):
    """Callback to check and prompt for file overwrite."""

    try:
        manifest = get_manifest(ctx)
        if not value:
            should_overwrite = click.confirm(
                f"Another ArkitektNext app {manifest.to_console_string()} exists already at {os.getcwd()}?. Do you want to overwrite?",
                abort=True,
            )
            if not should_overwrite:
                ctx.abort()
    except click.ClickException:
        pass

    return value


def ensure_semver(ctx, param, value):
    """Callback to check and prompt for file overwrite."""

    if not value:
        value = click.prompt(
            "The version of your app",
            default="0.0.1",
        )

    while not semver.Version.is_valid(value):
        get_console().print(
            "ArkitektNext versions need to follow [link=https://semver.org]semver[/link]. Please choose a correct format (examples: 0.0.0, 0.1.0, 0.0.0-alpha.1)"
        )
        value = click.prompt(
            "The version of your app",
            default="0.0.1",
        )

    return value


@click.command()
@click.option(
    "--overwrite-manifest",
    "-om",
    help="Should we overwrite the existing manifest if it already exists?",
    is_flag=True,
    default=False,
    callback=check_overwrite,
)
@click.option(
    "--template",
    "-t",
    help="The template to use. You can choose from a variety of preconfigured templates. They are just starting points and can be changed later.",
    type=click.Choice(compile_templates()),
    default="simple",
)
@click.option(
    "--identifier",
    "-i",
    help="The identifier of your app. This will be used to identify your app in the ArkitektNext ecosystem. It should be unique and should follow the [link=https://en.wikipedia.org/wiki/Reverse_domain_name_notation]reverse domain name notation[/link] (example: com.example.myapp)",
    prompt="Your app identifier",
    default=os.path.basename(os.getcwd()),
)
@click.option(
    "--version",
    "-v",
    help="The version of your app. Needs to follow [link=https://semver.org/]semantic versioning[/link].",
    default="0.0.1",
    callback=ensure_semver,
)
@click.option(
    "--author",
    help="The author of your app. This will be shown to users of your app",
    prompt="Your name",
    default=getuser(),
)
@click.option(
    "--template",
    help="Which template to use top create entrypoint",
    prompt="Your app file template",
    type=click.Choice(compile_templates()),
    default="simple",
)
@click.option(
    "--logo",
    help="Which logo to use for this app, needs to be a valid url",
    required=False,
)
@click.option(
    "--entrypoint",
    "-e",
    help="The entrypoint of your app. This will be the name of the python file. Omit the .py ending",
    prompt="Your app file",
    default="app",
)
@click.option(
    "--overwrite-app",
    "-oa",
    help="Do you want to overwrite the app file if it exists?",
    is_flag=True,
    default=False,
    callback=check_overwrite_app,
)
@click.option(
    "--scopes",
    "-s",
    help="The scopes of the app. You can choose multiple for your app. For a list of scopes, run `arkitekt_next manifest scopes available`",
    type=click.Choice(compile_scopes()),
    multiple=True,
    default=["read"],
)
@click.pass_context
def init(
    ctx,
    identifier: str,
    version: str,
    author: str,
    logo: Optional[str],
    scopes: List[str],
    template: str,
    entrypoint: str,
    overwrite_manifest: bool,
    overwrite_app: bool,
):
    """Initializes an ArkitektNext app

    This command will create a new ArkitektNext app in the current directory. It will
    create a `.arkitekt_next` folder that will contain a manifest and a `app.py` file,
    which will serve as the entrypoint for your app. By default, the app will be
    initialized with a simple hello world app, but you can choose from a variety
    of templates.

    """

    console = get_console(ctx)

    manifest = Manifest(
        logo=logo,
        author=author,
        identifier=identifier,
        version=version,
        scopes=scopes,
        entrypoint=entrypoint,
    )

    with open(build_relative_dir("templates", f"{template}.py")) as f:
        template_app = f.read()

    if not os.path.exists(f"{entrypoint}.py") or overwrite_app:
        with open(f"{entrypoint}.py", "w") as f:
            f.write(template_app)

    write_manifest(manifest)
    md = Panel(
        f"{manifest.to_console_string()} was successfully initialized\n\n"
        + "[not bold white]We are excited to see what you come up with!",
        border_style="green",
        style="green",
    )
    console.print(md)



import rich_click as click
from click import Context
from dokker.projects.contrib.konstruktor import KonstruktorProject, RemoteRepo
import asyncio
from typing import Optional
import os
from arkitekt_next.cli.vars import get_console

DEFAULT_REPO_URL = "https://arkitekt_next.live/repo.json"


@click.command()
@click.option(
    "--repo-url",
    help="The repo to use",
    default=DEFAULT_REPO_URL,
    prompt="What repo do you want to use?",
)
@click.option(
    "--channel",
    help="The channel to use",
    default="beta",
    prompt="What channel do you want to use?",
)
@click.option(
    "--name",
    help="The name of the deployment",
    default=None,
    required=False,
)
@click.option(
    "--overwrite",
    help="Should we overwrite the old deployment?",
    default=False,
    is_flag=True,
)
@click.option(
    "--setup",
    "-s",
    "setup",
    help="Additional setup variables",
    type=(str, str),
    multiple=True,
)
@click.pass_context
def init(
    ctx: Context,
    channel: str,
    repo_url: str,
    overwrite: bool = False,
    setup: Optional[list] = None,
    name: Optional[str] = None,
) -> None:
    """
    Initializes a new server

    """

    extra_content = dict(setup or [])

    project = KonstruktorProject(
        channel=channel,
        repo=RemoteRepo(url=repo_url),
        reinit_if_exists=overwrite,
        name=name,
        extra_context=extra_content,
    )

    try:
        asyncio.run(project.ainititialize())
        get_console(ctx).print("Done :)")
    except Exception as e:
        get_console(ctx).print_exception()
        raise click.ClickException("Failed to initialize project") from e



import rich_click as click
from click import Context
from dokker.projects.dokker import DokkerProject
from dokker.loggers.print import PrintLogger
from dokker.deployment import Deployment
from typing import Optional
import os
from .utils import compile_options


@click.command()
@click.option(
    "--name",
    help="The name of the deployment",
    default=None,
    required=False,
    type=click.Choice(compile_options()),
)
@click.pass_context
def remove(
    ctx: Context,
    name: Optional[str] = None,
) -> None:
    """
    Remove a deployment

    Removing a deployment will stop all containers and call docker compose down
    on the project. This will remove all containers and networks created by the
    deployment. As opposed to the down command, the deployment will also be
    removed from the .dokker folder, and all data stored in volumes managed by
    the deployment will be removed.

    This step is irreversible, so use with caution.

    """
    if not name:
        options = compile_options()
        if not options:
            raise click.ClickException(
                "No deployments found. Please run arkitekt_next server init first"
            )

        name = options[0]

    print(f"Running {name}")

    project = DokkerProject(
        name=name,
    )

    deployment = Deployment(project=project, logger=PrintLogger())

    with deployment:
        print("Shutting down...")
        deployment.down()

        print("Removing...")
        deployment.remove()

    print("Done")



"""
This module contains the types for the apps
depending on the builder used.

This module imports all the apps and their types
and sets them as attributes on the App class, if they are available.
If they are not available, they are set to Any, so that we can add
an import exception to the app.


"""

import logging
from typing import TYPE_CHECKING, Any, Dict
from arkitekt_next.model import Manifest
from koil.composition import Composition
from fakts import Fakts
from herre import Herre

logger = logging.getLogger(__name__)


class App(Composition):
    """An app that is built with the easy builder"""

    fakts: Fakts
    herre: Herre
    manifest: Manifest
    services: Dict[str, Any]

    def run(self):
        """Run the app"""
        self.services["rekuest"].run()

    async def __aenter__(self):
        await super().__aenter__()
        for service in self.services.values():
            await service.__aenter__()

        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        for service in self.services.values():
            await service.__aexit__(exc_type, exc_value, traceback)







import rich_click as click
from click import Context
from dokker.projects.dokker import DokkerProject
from dokker.loggers.print import PrintLogger
from dokker.deployment import Deployment
from typing import Optional
import os
from .utils import compile_options


@click.command()
@click.option(
    "--name",
    help="The name of the deployment",
    default=None,
    required=False,
    type=click.Choice(compile_options()),
)
@click.pass_context
def stop(
    ctx: Context,
    name: Optional[str] = None,
) -> None:
    """
    Stop a deployment

    Stopping a deployment will stop all containers and call docker compose stop
    on the project. This will stop all containers and networks created by the
    deployment. The deployment will still be available in the .dokker folder

    This should not be confused with the down command, which will remove all
    containers and remove all networks created by the deployment.

    """
    if not name:
        options = compile_options()
        if not options:
            raise click.ClickException(
                "No deployments found. Please run arkitekt_next server init first"
            )

        name = options[0]

    print(f"Stopping {name}")

    project = DokkerProject(
        name=name,
    )

    deployment = Deployment(project=project, logger=PrintLogger())

    with deployment:
        print("Shutting down...")
        deployment.stop()

    print("Done")



from importlib import import_module
from typing import Callable
from arkitekt_next.apps.types import App


def import_builder(builder: str) -> Callable[..., App]:
    module_path, function_name = builder.rsplit(".", 1)
    module = import_module(module_path)
    function = getattr(module, function_name)
    return function



import rich_click as click
from .init import init
from .up import up
from .down import down
from .inspect import inspect
from .remove import remove
from .open import open
from .stop import stop
from click import Context


@click.group()
@click.pass_context
def server(ctx: Context) -> None:
    """Manages your arkitekt_next deployments

    You can directly deploy your app to a server with the arkitekt_next server command. This will
    create a local docker-compose.yml file and a .arkitekt_next folder in your project. You can then
    use the arkitekt_next server build command to build your app for port and the arkitekt_next server run
    command to run your app on the server.

    """

    pass


server.add_command(init, "init")
server.add_command(up, "up")
server.add_command(down, "down")
server.add_command(remove, "remove")
server.add_command(inspect, "inspect")
server.add_command(open, "open")
server.add_command(stop, "stop")



import os
from typing import List


def compile_options() -> List[str]:
    """Compiles the available options for the server up and down commands:

    All deployments in the .arkitekt_next/deployments folder
    """
    deployments = os.path.join(os.getcwd(), ".dokker")

    available_deployments = []

    if not os.path.exists(deployments):
        return available_deployments

    for deployment in os.listdir(deployments):
        if os.path.isdir(os.path.join(deployments, deployment)):
            available_deployments.append(deployment)

    return available_deployments


def build_server_path(name: str) -> str:
    """Builds the path to the server

    Args:
        name (str): The name of the server

    Returns:
        str: The path to the server
    """
    return os.path.join(os.getcwd(), ".arkitekt_next", "servers", name)



import rich_click as click
from click import Context
from dokker.projects.dokker import DokkerProject
from dokker.loggers.print import PrintLogger
from dokker.deployment import Deployment
from typing import Optional
import os
from .utils import compile_options
from rich.table import Table
from rich.live import Live
from arkitekt_next.cli.vars import get_console
import json

DEFAULT_REPO_URL = (
    "https://raw.githubusercontent.com/jhnnsrs/konstruktor/master/repo/channels.json"
)


@click.command()
@click.option(
    "--name",
    help="The name of the deployment",
    default=None,
    required=False,
    type=click.Choice(compile_options()),
)
@click.pass_context
def inspect(
    ctx: Context,
    name: Optional[str] = None,
) -> None:
    """
    Inspect a deployment

    Inspect helps you find information about a deployment. It will show you
    information about the containers, networks and volumes created by the
    deployment. If you have not run arkitekt_next server init before, this will
    fail.

    """
    if not name:
        options = compile_options()
        if not options:
            raise click.ClickException(
                "No deployments found. Please run arkitekt_next server init first"
            )

        name = options[0]

    project = DokkerProject(
        name=name,
    )

    console = get_console(ctx)

    deployment = Deployment(project=project)

    with deployment:
        print(json.dumps(deployment.inspect().dict(), indent=4))



import rich_click as click
from click import Context
from dokker.projects.dokker import DokkerProject
from dokker.loggers.print import PrintLogger
from dokker.deployment import Deployment
from typing import Optional
import os
from .utils import compile_options


@click.command()
@click.option(
    "--name",
    help="The name of the deployment",
    default=None,
    required=False,
    type=click.Choice(compile_options()),
)
@click.pass_context
def down(
    ctx: Context,
    name: Optional[str] = None,
) -> None:
    """
    Down a deployment

    Removing a deployment will stop all containers and call docker compose down
    on the project. This will remove all containers and networks created by the
    deployment. The deployment will still be available in the .dokker folder
    However (depending on your settings) data that was stored in volumes managed
    by the deployment will be removed.

    If you want to simple Stop the deployment, use the stop command instead.

    """
    if not name:
        options = compile_options()
        if not options:
            raise click.ClickException(
                "No deployments found. Please run arkitekt_next server init first"
            )

        name = options[0]

    print(f"Running {name}")

    project = DokkerProject(
        name=name,
    )

    deployment = Deployment(project=project, logger=PrintLogger())
    with deployment:
        deployment.down()
        print("Shutting down...")

    print("Done")



from pydantic import BaseModel, Field
from herre import Herre
from fakts import Fakts
from .model import Manifest, Requirement
from typing import Callable, Dict
import importlib
import sys
import os
import traceback
import logging
import pkgutil

Params = Dict[str, str]


class Registration(BaseModel):
    name: str
    requirement: Requirement
    builder: Callable[[Herre, Fakts, Params], object]


basic_requirements = {
    "lok": Requirement(
            service="live.arkitekt.lok",
            description="An instance of ArkitektNext Lok to authenticate the user",
        ),
}


class ServiceBuilderRegistry:
    def __init__(self):
        self.service_builders = {}
        self.requirements = basic_requirements

    def register(self, name: str, service_builder: Callable[[Herre, Fakts], object], requirement: Requirement):
        self.service_builders[name] = service_builder
        self.requirements[name] = requirement

    def get(self, name):
        return self.services.get(name)
    

    def build_service_map(self, fakts: Fakts, herre: Herre, params: Params, manifest: Manifest):
        return {name: builder(fakts, herre, params, manifest) for name, builder in self.service_builders.items()}
    
    def get_requirements(self):
        return self.requirements
    





class SetupInfo:
    services: Dict[str, object]



def check_and_import_services() -> ServiceBuilderRegistry:

    service_builder_registry =  ServiceBuilderRegistry()
    
    # Function to load and call init_extensions from __rekuest__.py
    def load_and_call_init_extensions(module_name, rekuest_path):
        try:
            spec = importlib.util.spec_from_file_location(f"{module_name}.__arkitekt__", rekuest_path)
            rekuest_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(rekuest_module)
            if hasattr(rekuest_module, 'init_services'):
                rekuest_module.init_services(service_builder_registry)
                logging.info(f"Called init_service function from {module_name}")
            else:
                print(f"No init_services function in {module_name}.__arkitekt__")
        except Exception as e:
            print(f"Failed to call init_services for {module_name}: {e}")
            traceback.print_exc()

    # Check local modules in the current working directory
    current_directory = os.getcwd()
    for item in os.listdir(current_directory):
        item_path = os.path.join(current_directory, item)
        if os.path.isdir(item_path) and os.path.isfile(os.path.join(item_path, '__init__.py')):
            rekuest_path = os.path.join(item_path, '__arkitekt__.py')
            if os.path.isfile(rekuest_path):
                load_and_call_init_extensions(item, rekuest_path)

    # Check installed packages
    for _, module_name, _ in pkgutil.iter_modules():
        try:
            module_spec = importlib.util.find_spec(module_name)
            if module_spec and module_spec.origin:
                rekuest_path = os.path.join(os.path.dirname(module_spec.origin), '__arkitekt__.py')
                if os.path.isfile(rekuest_path):
                    load_and_call_init_extensions(module_name, rekuest_path)
        except Exception as e:
            print(f"Failed to call init_extensions for installed package {module_name}: {e}")
            traceback.print_exc()

    return service_builder_registry



import rich_click as click

from arkitekt_next.constants import DEFAULT_ARKITEKT_URL
from .constants import *
from .types import *
from .vars import *
from .ui import *

with_fakts_url = click.option(
    "--url",
    "-u",
    help="The fakts url for connection",
    default=DEFAULT_ARKITEKT_URL,
    envvar="FAKTS_URL",
)

with_token = click.option(
    "--token",
    "-t",
    help="The token for the fakts instance",
    envvar="FAKTS_TOKEN",
    required=False,
)
with_version = click.option(
    "--version",
    "-v",
    help="Override the version of the app",
    envvar="ARKITEKT_VERSION",
)

with_log_level = click.option(
    "--log",
    "-l",
    help="Override the logging level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    envvar="ARKITEKT_LOG_LEVEL",
)


with_skip_cache = click.option(
    "--no-cache",
    "-nc",
    is_flag=True,
    default=False,
    help="Should we skip the cache",
    envvar="ARKITEKT_NO_CACHE",
)

with_instance_id = click.option(
    "--instance-id",
    "-i",
    default="main",
    help="The token for the fakts instance",
    envvar="REKUEST_INSTANCE",
)

with_log_level = click.option(
    "--log-level",
    "-l",
    default="ERROR",
    help="The token for the fakts instance",
    envvar="ARKITEKT_LOG_LEVEL",
)


with_builder = click.option(
    "--builder",
    "-b",
    default="arkitekt_next.builders.easy",
    help="The builder for this run",
    envvar="ARKITEKT_BUILDER",
)

with_headless = click.option(
    "--headless",
    "-h",
    is_flag=True,
    default=False,
    help="Should we start headless",
    envvar="ARKITEKT_HEADLESS",
)


def check_gen_boring(ctx, param, value):
    """Callback to check and prompt for file overwrite."""

    if not value:
        get_console(ctx).print(construct_codegen_welcome_panel())

    return value


with_boring = click.option(
    "--boring",
    help="Should we skip the welcome message?",
    is_flag=True,
    default=False,
    callback=check_gen_boring,
)
with_seperate_document_dirs = click.option(
    "--seperate-doc-dirs",
    "-sd",
    help="Should we generate seperate dirs for the documents?",
    is_flag=True,
    default=False,
)
with_choose_services = click.option(
    "--services",
    "-s",
    help="The services to create the codegen for",
    multiple=True,
    type=click.Choice(compile_services()),
    default=["mikro"],
)
with_graphql_config = click.option(
    "--config",
    "-c",
    help="The name of the configuration file",
    type=str,
    default="graphql.config.yaml",
)
with_api_path = click.option(
    "--path",
    "-c",
    help="The path of the api to be generated",
    prompt="Where should we generate the api? (relative to the current directory)",
    type=str,
    default="api",
)


def check_overwrite_config(ctx, param, value):
    """Callback to check and prompt for file overwrite."""

    config = ctx.params["config"]
    if os.path.exists(config) and not value:
        should_overwrite = click.confirm(
            "GraphQL Config file already exists. Do you want to overwrite?"
        )
        return should_overwrite

    return value


with_overwrite_graphql = click.option(
    "--overwrite-config",
    "-o",
    help="Should we overwrite the config file if it already exists",
    is_flag=True,
    default=False,
    callback=check_overwrite_config,
)
with_documents = click.option(
    "--documents",
    "-d",
    help="With documents",
    is_flag=True,
    default=True,
)
with_schemas = click.option(
    "--schemas",
    "-s",
    help="Should we copy the schemas",
    is_flag=True,
    default=True,
)



import sys
import os

try:
    import rich_click as click

    from rich.console import Console
except ImportError:
    print(
        "ArkitektNext CLI is not installed, please install it first. By installing the cli, e.g with `pip install arkitekt_next[cli]`, you can use the `arkitekt_next` command."
    )
    sys.exit(1)

from arkitekt_next.cli.vars import *
from arkitekt_next.cli.constants import *
from arkitekt_next.cli.texts import *
from arkitekt_next.cli.commands.run.main import run
from arkitekt_next.cli.commands.gen.main import gen
from arkitekt_next.cli.commands.server.main import server
from arkitekt_next.cli.commands.kabinet.main import kabinet
from arkitekt_next.cli.commands.init.main import init
from arkitekt_next.cli.commands.manifest.main import manifest
from arkitekt_next.cli.commands.inspect.main import inspect
from arkitekt_next.cli.commands.call.main import call
from arkitekt_next.cli.io import load_manifest
from arkitekt_next.utils import create_arkitekt_next_folder

default_docker_file = """
FROM python:3.8-slim-buster


RUN pip install arkitekt_next==0.4.23


RUN mkdir /app
COPY . /app
WORKDIR /app

"""


click.rich_click.HEADER_TEXT = LOGO
click.rich_click.ERRORS_EPILOGUE = ERROR_EPILOGUE
click.rich_click.USE_RICH_MARKUP = True


@click.group()
@click.pass_context
def cli(ctx):
    """ArkitektNext is a framework for building safe and performant apps that then can be centrally orchestrated and managed
    in workflows.


    This is the CLI for the ArkitektNext Python SDK. It allows you to create and deploy ArkitektNext Apps from your python code
    as well as to run them locally for testing and development. For more information about ArkitektNext, please visit
    [link=https://arkitekt_next.live]https://arkitekt_next.live[/link]
    """
    sys.path.append(os.getcwd())

    ctx.obj = {}
    console = Console()
    set_console(ctx, console)

    create_arkitekt_next_folder()

    manifest = load_manifest()
    if manifest:
        set_manifest(ctx, manifest)

    pass


cli.add_command(init, "init")
cli.add_command(run, "run")
cli.add_command(gen, "gen")
cli.add_command(kabinet, "kabinet")
cli.add_command(manifest, "manifest")
cli.add_command(inspect, "inspect")
cli.add_command(server, "server")
cli.add_command(call, "call")

if __name__ == "__main__":
    cli()



import rich_click as click
from click import Context
from dokker.projects.dokker import DokkerProject
from dokker.loggers.print import PrintLogger
from dokker.deployment import Deployment
from typing import Optional
import os
from .utils import compile_options
from rich.table import Table
from rich.live import Live
from arkitekt_next.cli.vars import get_console


DEFAULT_REPO_URL = (
    "https://raw.githubusercontent.com/jhnnsrs/konstruktor/master/repo/channels.json"
)


@click.command()
@click.option(
    "--name",
    help="The name of the deployment",
    default=None,
    required=False,
    type=click.Choice(compile_options()),
)
@click.pass_context
def up(
    ctx: Context,
    name: Optional[str] = None,
) -> None:
    """
    Ups a deployment

    Up starts your deployment. This will create all containers and networks
    needed for your deployment. If you have not run arkitekt_next server init
    before, this will fail.


    """
    if not name:
        options = compile_options()
        if not options:
            raise click.ClickException(
                "No deployments found. Please run arkitekt_next server init first"
            )

        name = options[0]

    print(f"Running {name}")

    project = DokkerProject(
        name=name,
    )

    console = get_console(ctx)

    logger = PrintLogger(print_function=lambda x: console.log(x))

    deployment = Deployment(project=project, logger=logger)

    with deployment:
        port = deployment.inspect().find_service("lok").get_label("arkitekt_next.link")

        status = console.status(
            f"[bold green]Running on the Lok Server Port {port} Press Ctrl+c once to cancel..."
        )

        with status:
            x = deployment.up(detach=False)



"""Models for ArkitektNext. Thiese include extensiosn for the Fakts Manifest and the User model."""

from hashlib import sha256
from pydantic import BaseModel, Field
from typing import List, Optional


class Requirement(BaseModel):
    service: str
    """ The service is the service that will be used to fill the key, it will be used to find the correct instance. It needs to fullfill
    the reverse domain naming scheme"""
    optional: bool = False
    """ The optional flag indicates if the requirement is optional or not. Users should be able to use the client even if the requirement is not met. """
    description: Optional[str] = None
    """ The description is a human readable description of the requirement. Will be show to the user when asking for the requirement."""


def build_default_requirements() -> dict[str, Requirement]:
    return {
        "lok": Requirement(
            service="live.arkitekt_next.lok",
            description="An instance of ArkitektNext Lok to authenticate the user",
        ),
        "rekuest": Requirement(
            service="live.arkitekt_next.rekuest",
            description="An instance of ArkitektNext Rekuest to assign to nodes",
        ),
        "kabinet": Requirement(
            service="live.arkitekt_next.kabinet",
            description="An instance of ArkitektNext Kabinet to retrieve nodes from",
        ),
        "mikro": Requirement(
            service="live.arkitekt_next.mikro",
            description="An instance of ArkitektNext Mikro to make requests to the user's data",
            optional=True,
        ),
        "fluss": Requirement(
            service="live.arkitekt_next.fluss",
            description="An instance of ArkitektNext Fluss to make requests to the user's data",
            optional=False,
        ),
        "port": Requirement(
            service="live.arkitekt_next.port",
            description="An instance of ArkitektNext Fluss to make requests to the user's data",
            optional=True,
        ),
        "datalayer": Requirement(
            service="live.arkitekt_next.datalayer",
            description="An instance of ArkitektNext Datalayer to make requests to the user's data",
            optional=False,
        ),
    }


class Manifest(BaseModel):
    """A manifest for an app that can be installed in ArkitektNext

    Manifests are used to describe apps that can be installed in ArkitektNext.
    They provide information about the app, such as the
    its globally unique identifier, the version, the scopes it needs, etc.

    This Manifest is send to the Fakts server on initial app configuration,
    and is used to register the app with the Fakts server, which in turn
    will prompt the user to grant the app access to establish itself as
    an ArkitektNext app (and therefore as an OAuth2 client) (see more in the
    Fakts documentation).

    """

    version: str
    """ The version of the app TODO: Should this be a semver? """
    identifier: str
    """ The globally unique identifier of the app: TODO: Should we check for a reverse domain name? """
    scopes: List[str]
    """ Scopes that this app should request from the user """
    logo: Optional[str]
    """ A URL to the logo of the app TODO: We should enforce this to be a http URL as local paths won't work """
    requirements: Optional[dict[str, Requirement]] = Field(
        default_factory=build_default_requirements
    )
    """ Requirements that this app has TODO: What are the requirements? """

    class Config:
        extra = "forbid"


    def hash(self):
        """Hash the manifest"""
        return sha256(self.json(sort_keys=True).encode()).hexdigest()

class User(BaseModel):
    """A user of ArkitektNext

    This model represents a user on ArkitektNext. As herre is acgnostic to the
    user model, we need to provide a model that can be used to represent
    the ArkitektNext user. This model is used by the
    :class:`herre.fakts.fakts_endpoint_fetcher.FaktsUserFetcher` to
    fetch the user from the associated ArkitektNext Lok instance. This model
    is closely mimicking the OIDC user model, and is therefore compatible
    to represent OIDC users.

    """

    id: str = Field(alias="sub")
    """ The user's id (in lok, this is the user's sub(ject) ID)"""

    username: str = Field(alias="preferred_username")
    """ The user's preferred username """
    email: str = Field(alias="email")
    """ The user's preferred username """
